package meerkat.mixer.proofs;

import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.crypto.concrete.ECElGamalEncryption;
import meerkat.crypto.mixnet.Mix2ZeroKnowledgeVerifier;
import meerkat.protobuf.Crypto;
import meerkat.protobuf.Mixing;
import org.bouncycastle.math.ec.ECPoint;
import org.factcenter.qilin.primitives.RandomOracle;
import org.factcenter.qilin.primitives.concrete.ECGroup;

/**
 * implements Mix2ZeroKnowledgeVerifier
 */
public class Verifier implements Mix2ZeroKnowledgeVerifier {


    private final ECGroup group;
    private final RandomOracle randomOracle;
    private final ECPoint g,h;
    private final ECElGamalMixProof organizer;
    private final ZeroKnowledgeOrProofParser parser;

    /**
     * constructor
     * @param encryptor should be as the encryptor used by meerkat.mixer.proofs
     * @param randomOracle should be as the random oracle used by meerkat.mixer.proofs
     */
    public Verifier(ECElGamalEncryption encryptor, RandomOracle randomOracle) {
        this.group = encryptor.getGroup();
        this.g = group.getGenerator();
        this.h = encryptor.getElGamalPK().getPK();
        this.randomOracle = randomOracle;
        this.organizer = new ECElGamalMixProof(group,g,h);
        this.parser = new ZeroKnowledgeOrProofParser(group);
    }

    /**
     * verify zero knowledge proof
     * @param in1
     * @param in2
     * @param out1
     * @param out2
     * @param proof out1 = rerandomize(in1) && out2 = rerandomize(in2)
     *              ||
     *              out1 = rerandomize(in2) && out2 = rerandomize(in1)
     * @return  true iff all 4 or proofs are valid
     * @throws InvalidProtocolBufferException
     */
    public boolean verify(Crypto.RerandomizableEncryptedMessage in1,
                          Crypto.RerandomizableEncryptedMessage in2,
                          Crypto.RerandomizableEncryptedMessage out1,
                          Crypto.RerandomizableEncryptedMessage out2,
                          Mixing.ZeroKnowledgeProof proof) throws InvalidProtocolBufferException {
        ECElGamalMixProof.Statement statement = organizer.createProofInput(in1,in2,out1,out2);
        return verifyElGamaOrProof(statement.getFirst(), proof.getFirst())&&
                verifyElGamaOrProof(statement.getSecond(), proof.getSecond())&&
                verifyElGamaOrProof(statement.getThird(), proof.getThird())&&
                verifyElGamaOrProof(statement.getFourth(), proof.getFourth());

    }


    /**
     *
     * @param orStatement
     * @param orProof
     * @return verify single or proof
     */
    private boolean verifyElGamaOrProof(ECElGamalMixProof.OrStatement orStatement,
                                        Mixing.ZeroKnowledgeProof.OrProof orProof) {
        ZeroKnowledgeOrProofParser.ZeroKnowledgeOrProofContainer container = parser.parseOrProof(orProof);
        return  container.g1.equals(orStatement.g1)       &&
                container.h1.equals(orStatement.h1)       &&
                container.g2.equals(orStatement.g2)       &&
                container.h2.equals(orStatement.h2)       &&
                container.g1Tag.equals(orStatement.g1Tag) &&
                container.h1Tag.equals(orStatement.h1Tag) &&
                container.g2Tag.equals(orStatement.g2Tag) &&
                container.h2Tag.equals(orStatement.h2Tag) &&
                container.c1.add(container.c2).mod(group.orderUpperBound())
                        .equals(Prover.hash(container.forRandomOracle,randomOracle).mod(group.orderUpperBound())) &&
                group.multiply(container.g1, container.z)
                        .equals(group.add(container.u, group.multiply(container.h1,container.c1))) &&
                group.multiply(container.g2, container.z)
                        .equals(group.add(container.v, group.multiply(container.h2,container.c1))) &&
                group.multiply(container.g1Tag, container.zTag)
                        .equals(group.add(container.uTag, group.multiply(container.h1Tag,container.c2))) &&
                group.multiply(container.g2Tag, container.zTag)
                        .equals(group.add(container.vTag, group.multiply(container.h2Tag,container.c2)));

    }
}
