package meerkat.mixer;

import com.google.protobuf.ByteString;
import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.crypto.concrete.ECElGamalEncryption;
import meerkat.crypto.mixnet.Mix2ZeroKnowledgeProver;
import meerkat.crypto.mixnet.Mix2ZeroKnowledgeVerifier;
import meerkat.mixer.proofs.Prover;
import meerkat.mixer.proofs.Verifier;
import meerkat.protobuf.ConcreteCrypto;
import meerkat.protobuf.Crypto;
import meerkat.protobuf.Voting;
//import meerkat.protobuf.Voting.PlaintextBallot;
import org.bouncycastle.math.ec.ECPoint;
import org.factcenter.qilin.primitives.RandomOracle;
import org.factcenter.qilin.primitives.concrete.DigestOracle;
import org.factcenter.qilin.primitives.concrete.ECElGamal;
import org.factcenter.qilin.primitives.concrete.ECGroup;
import org.junit.Before;
import org.junit.Test;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.math.BigInteger;
import java.util.Random;

/**
 * Created by Tzlil on 12/31/2015.
 */
public class ZeroKnowledgeProofTest {

    Random rand;
    ECElGamal.SK key;
    ECGroup group;
    ECElGamalEncryption enc;
    ConcreteCrypto.ElGamalPublicKey serializedPk;
    Mix2ZeroKnowledgeVerifier verifier ;
    Mix2ZeroKnowledgeProver prover ;

    @Before
    public void setup() throws Exception {
        rand = new Random();
        group = new ECGroup("secp256k1");
        BigInteger sk = ECElGamal.generateSecretKey(group, rand);
        key = new ECElGamal.SK(group, sk);
        serializedPk = Utils.serializePk(group, key);
        enc = new ECElGamalEncryption();
        enc.init(serializedPk);
        RandomOracle randomOracle = new DigestOracle();
        verifier = new Verifier(enc,randomOracle);
        prover = new Prover(new Random(),enc,randomOracle);
    }

    private ECPoint convert2ECPoint(ByteString bs){
        return group.decode(bs.toByteArray());
    }

    public void oneZKPTest() throws InvalidProtocolBufferException {
        Voting.PlaintextBallot msg1 = Utils.genRandomBallot(2,3,16); // 2 questions with 3 answers each, in range 0-15.
        Voting.PlaintextBallot msg2 = Utils.genRandomBallot(2,3,16);
        Crypto.EncryptionRandomness r1 = enc.generateRandomness(rand);
        Crypto.EncryptionRandomness r2 = enc.generateRandomness(rand);

        Crypto.RerandomizableEncryptedMessage e1 = enc.encrypt(msg1, enc.generateRandomness(rand));
        Crypto.RerandomizableEncryptedMessage e2 = enc.encrypt(msg2, enc.generateRandomness(rand));
        Crypto.RerandomizableEncryptedMessage e1New = enc.rerandomize(e1, r1);
        Crypto.RerandomizableEncryptedMessage e2New = enc.rerandomize(e2, r2);

        assertEquals (Utils.decrypt(Voting.PlaintextBallot.class, key, group, e1),    msg1);
        assertEquals (Utils.decrypt(Voting.PlaintextBallot.class, key, group, e1New), msg1);
        assertEquals (Utils.decrypt(Voting.PlaintextBallot.class, key, group, e2),    msg2);
        assertEquals (Utils.decrypt(Voting.PlaintextBallot.class, key, group, e2New), msg2);

        ECPoint g = group.getGenerator();
        ECPoint h = enc.getElGamalPK().getPK();

        ConcreteCrypto.ElGamalCiphertext e1ElGamal = ECElGamalEncryption.RerandomizableEncryptedMessage2ElGamalCiphertext(e1);
        ConcreteCrypto.ElGamalCiphertext e2ElGamal = ECElGamalEncryption.RerandomizableEncryptedMessage2ElGamalCiphertext(e2);
        ConcreteCrypto.ElGamalCiphertext e1TagElGamal = ECElGamalEncryption.RerandomizableEncryptedMessage2ElGamalCiphertext(e1New);
        ConcreteCrypto.ElGamalCiphertext e2TagElGamal = ECElGamalEncryption.RerandomizableEncryptedMessage2ElGamalCiphertext(e2New);


        assertEquals (g.multiply(enc.extractRandomness(r1)),
                group.add(convert2ECPoint(e1TagElGamal.getC1()),group.negate(convert2ECPoint(e1ElGamal.getC1()))));
        assertEquals (h.multiply(enc.extractRandomness(r1)),
                group.add(convert2ECPoint(e1TagElGamal.getC2()),group.negate(convert2ECPoint(e1ElGamal.getC2()))));
        assertEquals (g.multiply(enc.extractRandomness(r2)),
                group.add(convert2ECPoint(e2TagElGamal.getC1()),group.negate(convert2ECPoint(e2ElGamal.getC1()))));
        assertEquals (h.multiply(enc.extractRandomness(r2)),
                group.add(convert2ECPoint(e2TagElGamal.getC2()),group.negate(convert2ECPoint(e2ElGamal.getC2()))));

        assertTrue (verifier.verify(e1,e2,e1New,e2New,prover.prove(e1,e2,e1New,e2New,false,0,0,0,r1,r2)));
    }

    @Test
    public void zeroKnowledgeProofTest() throws InvalidProtocolBufferException {

        int tests = 1000;

        for (int i = 0; i < tests; i ++){
            System.out.println("ZKP test #" + i);
            oneZKPTest();
        }
    }
}
